import os,sys
import re,bisect,copy
import numpy as np
import collections

from rasmus import util

from yjw.ordereddict import *

# ==========================================================
# Architecture data structures

class MultiSetArch:
    """
    list of architectures (stored as sorted list)
    architectures are stored in hyphenated form
    """
    
    def __init__(self,lst=[]):
        self.__lst = sorted(lst)

    def add_arch(self,arch):
        bisect.insort(self.__lst,arch)

    def remove_arch(self,arch):
        self.__lst.remove(arch)

    def archs(self):
        return self.__lst

    def regions(self):
        regs = []
        for arch in self.__lst:
            regs.extend(arch.split('-'))
        return regs

    def digraph(self):
        return MultiSetArchGraph(self.__lst)

    def add(self,other):
        """return NEW list"""
        return MultiSetArch(self.__lst+other.archs())

    def tostring(self):
        return ','.join([arch for arch in self.__lst])

    def deepcopy(self):
        # okay to use initializer since uses sorted list
        return MultiSetArch(self.__lst)


class MultiSetArchGraph:
    """
    a "forest" of directed graphs, where each graph represents an architecture
    """
    class NodeError(Exception):
        """
        Exception raised for error in MultiSetArchGraph.
        """
        pass

    class EdgeError(Exception):
        """
        Exception raised for error in MultiSetArchGraph.
        """
        pass
    
    def __init__(self,lst=None,other=None):
        import networkx as nx
        
        # either initialize (with list) or copy (with uid,DG,data)
        assert(not ((lst is not None) and (other is not None))),\
               "Can only specify lst or other."

        if other is not None:
            self.__uid = other.__uid
            self.__DG = nx.DiGraph()
            self.__DG.add_nodes_from(other.__DG.nodes())
            self.__DG.add_edges_from(other.__DG.edges())
            self.__data = {}
            for key,val in other.__data.iteritems():
                self.__data[key] = val
            return
        
        self.__uid = 0            # unique id to assign to next node added
        self.__DG = nx.DiGraph()  # unique ids are assigned to each node
                                  # no data is stored in edges
        self.__data = {}          # map of uid to reg
        if lst is None:
            return
        for arch in lst:
            self.add_arch(arch)

    def add_arch(self,arch):
        # return uids assigned to regions in arch
        regs = arch.split('-')
        nodes = range(self.__uid,self.__uid+len(regs))
        self.__DG.add_nodes_from(nodes)
        self.__DG.add_path(nodes)
        self.__data.update(dict(zip(nodes,regs)))
        self.__uid += len(regs)
        return nodes

    def add_region(self,reg):
        # return uid assigned to reg
        self.__DG.add_node(self.__uid)
        self.__data[self.__uid] = reg
        self.__uid += 1
        return self.__uid-1

    def remove_region(self,uid):
        # return reg assigned with uid
        if uid not in self.__data:
            raise self.NodeError("Node %s (%s) must alread be in graph" %\
                                 (uid,self.__data[uid]))
        if self.__DG.degree(uid) != 0:
            raise self.NodeError("Node %s (%s) must not be connected in graph" %\
                                 (uid,self.__data[uid]))
        self.__DG.remove_node(uid)
        reg = self.__data[uid]
        self.__data.__delitem__(uid)
        return reg

    def add_edge(self,uid,vid):
        # return regs assigned to uid,vid
        if uid == vid:
            raise self.EdgeError("Node %s and node %s must not be the same" %\
                                 (uid,vid))
        if not (uid in self.__data and vid in self.__data):
            raise self.NodeError("Nodes %s (%s) and %s (%s) must already be in graph" %\
                                 (uid,self.__data[uid],vid,self.__data[vid]))
        if self.__DG.out_degree(uid) != 0:
            other = list(self.__DG.out_degree_iter(uid))[0]
            raise self.EdgeError("Node %s (%s) already has a right node %s (%s)" %\
                                 (uid,self.__data[uid],other,self.__data[other]))
        if self.__DG.in_degree(vid) != 0:
            other = list(self.__DG.in_degree_iter(vid))[0]
            raise self.EdgeError("Node %s (%s) already has a left node %s (%s)" %\
                                 (vid,self.__data[vid],other,self.__data[other]))
        self.__DG.add_edge(uid,vid)
        return self.__data[uid],self.__data[vid]

    def remove_edge(self,uid,vid):
        # return regs assigned to uid,vid
        if not (uid in self.__data and vid in self.__data):
            raise self.NodeError("Nodes %s and %s must already be in graph" %\
                                 (uid,vid))
        if vid not in self.__DG[uid]:
            raise self.EdgeError("Edge %s (%s) - %s (%s) must already be in graph" %\
                                 (uid,self.__data[uid],vid,self.__data[vid]))
        self.__DG.remove_edge(uid,vid)
        return self.__data[uid],self.__data[vid]
        
    def edges(self,nbunch=None):
        # return edges with nbunch (uids)
        return self.__DG.edges(nbunch)

    def __right(self,dg,nbunch=None,with_labels=False):
        # return uids to the right (or None) of nbunch (uids) in the dg
        if not hasattr(nbunch,'__iter__'):
            r = dg[nbunch].keys()
            if with_labels:
                return {nbunch:r[0] if len(r)>0 else None}
            return r[0] if len(r)>0 else None
            
        d = dict(dg.edges(nbunch))
        if with_labels:
            return dict([(n,d.get(n,None)) for n in nbunch])
        return [d.get(n,None) for n in nbunch]

    def left(self,nbunch=None,with_labels=False):
        # return uids to the left (or None) of nbunch (uids)
        return self.__right(self.__DG.reverse(),nbunch,with_labels)
    
    def right(self,nbunch=None,with_labels=False):
        # return uids to the right (or None) of nbunch (uids)
        return self.__right(self.__DG,nbunch,with_labels)

    def regions(self):
        # return dict of (uid,region)
        return self.__data

    def archs(self,nbunch=None,with_uids=True,format='tuple'):
        # return lst of (arch strings) or list of (arch uids,arch strings)
        # where arch uid and arch string are in tuple format and 
        # where the arch contains at least one node in nbunch (uids)
        assert(format == 'tuple' or format == 'str')
        func = lambda it:tuple(it) \
               if format == 'tuple' \
               else '-'.join(map(str,it))

        if nbunch is None:
            nbunch = self.__data
        elif not hasattr(nbunch,'__iter__'):
            nbunch = [nbunch]

        archs = []
        rev = self.__DG.reverse()
        for uid in self.__data:
            if len(rev[uid]) == 0:
                # no left edge (e.g. start of arch) so follow the path
                u = [uid]
                s = [self.__data[uid]]
                next = self.__DG[uid].keys()
                while len(next) > 0:
                    vid = next[0]
                    u.append(vid)
                    s.append(self.__data[vid])
                    next = self.__DG[vid].keys()

                # keep if overlaps with nbunch
                if set(u).intersection(nbunch):
                    if with_uids:
                        archs.append((func(u),func(s)))
                    else:
                        archs.append(func(s))
        return archs

    def uids(self):
        # return dict of (region,list of uids)
        d = collections.defaultdict(list)
        for uid,reg in self.__data.iteritems():
            d[reg].append(uid)
        return d

    def deepcopy(self):
        # TODO: make efficient
        #return copy.deepcopy(self)
        return MultiSetArchGraph(other=self)

    def nodesequal(self,other):
        # node equality (but not uids)
        u = self.uids()
        v = other.uids()
        if len(u)!=len(v):
            return False
        for reg in u.iterkeys():
            if len(u[reg])!=len(v[reg]):
                return False
        return True

    def equal(self,other):
        # node and edge equality (but not uids)
        # really inefficient
        return cmp(self.tostring(),other.tostring())

    def tostring(self,with_uids=False):
        # string representation
        inp = sorted(self.archs(with_uids=with_uids,format='str'))
        if with_uids:
            return ','.join(['%s:%s' % (a[0],a[1]) for a in inp])
        return ','.join(inp)


# ==========================================================
# private utilities
def __findmatch(pattern,string,flags=0,errorstr=""):
    m = re.match(pattern,string,flags)
    if m == None:
        raise Exception(errorstr + string)
    else:
        return m.groups()

# ==========================================================
exparch = r"\[([a-zA-Z0-9.-]+)\]"
expgnarch = r"(\w+)" + exparch
expgnarch1 = r'"([^"]+)"' + exparch
expedge = r'\((\w+),(\w+)\)'
expevent = r'(\w+):([a-zA-Z0-9.,-]*)>([a-zA-Z0-9.,-]*)'
def parsearch(string):
    if '"' in string:
        return __findmatch(expgnarch1,string,errorstr="Invalid architecture: ")
    else:
        try:
            return __findmatch(expgnarch,string,errorstr="Invalid architecture: ")
        except:
            return (None,) + __findmatch(exparch,string,errorstr="Invalid architecture: ")
def parseedge(string):
    return __findmatch(expedge,string,errorstr="Invalid edge: ")
def parseevent(string):
    return __findmatch(expevent,string,errorstr="Invalid event: ")

def stringarch(arch):
    string = ''
    if arch[0] is not None:
        string += arch[0]
    string += '[%s]' % arch[1]
    return string
def stringevent(event):
    return '%s:%s>%s' % (event[0],event[1],event[2])

# ==========================================================
class ArchFam:
    """
    storage for architecture family
    """
    def __init__(self,inputs=None,tied=None,archs=None,events=None,costs=None):
        if inputs is None:
            self.inputs = OrderedDict()
        if tied is None:
            self.tied = OrderedDict()
        if archs is None:
            self.archs = OrderedDict()
        if events is None:
            self.events = OrderedDict()
        if costs is None:
            self.costs = OrderedDict()

    class StateError(Exception):
        """
        Raised when an operation violates the internal files state.

        Attributes:
            prev -- state at beginning of transition
            next -- attempted new state
            msg  -- explanation of why the specific transition is not allowed
        """

        def __init__(self, msg):
            self.msg = msg

        def __str__(self):
            return self.msg

    
    # states, state strings, and valid state transitions
    EMPTY,ARCHTREE,INPUTS,TIED,ARCHS,EVENTS,COSTS = STATES = range(7)
    __statestr = {EMPTY:   'EMPTY',
                  ARCHTREE:'ARCHTREE',
                  INPUTS:  'INPUTS',
                  TIED:    'TIED',
                  ARCHS:   'ARCHS',
                  EVENTS:  'EVENTS',
                  COSTS:   'COSTS'}
    __trans = {EMPTY:   [ARCHTREE],
               ARCHTREE:[EMPTY,INPUTS,TIED,ARCHS,EVENTS,COSTS],
               INPUTS:  [ARCHTREE],
               TIED:    [ARCHTREE],
               ARCHS:   [ARCHTREE],
               EVENTS:  [ARCHTREE],
               COSTS:   [ARCHTREE]}
    def __startstr(self,state):
        if state not in self.STATES:
            raise self.StateError("invalid state:", state)
        return '# %s' % self.__statestr[state]
    def __endstr(self,state):
        if state not in self.STATES:
            raise self.StateError("invalid state:", state)
        return '# END%s' % self.__statestr[state]
    def __statetrans(self,oldstate,newstate):
        if oldstate not in self.STATES:
            raise self.StateError("invalid state:", oldstate)
        if newstate not in self.STATES:
            raise self.StateError("invalid state:", newstate)
        if newstate not in self.__trans[oldstate]:
            raise self.StateError("invalid state transition: " +\
                                  self.__statestr[oldstate] + '->' +\
                                  self.__statestr[newstate])
        return True

    def add_inputs(self,odict):
        for key,val in odict.iteritems():
            self.add_input(key,val)
    def add_input(self,key,val):
        try:
            float(val)
        except ValueError:
            raise ValueError('Value not a valid number: ' + val)
        if key in self.inputs:
            raise Exception('Key already found in inputs: ' + key)
        self.inputs[key] = val
    
    def add_tieds(self,odict):
        for key,val in odict.iteritems():
            self.add_tied(key,val)
    def add_tied(self,key,val):
        assert(isinstance(val,bool))
        if key in self.tied:
            raise Exception('Key already found in tied: ' + key)
        self.tied[key] = val
    
    def add_arch(self,sp,arch):
        self.archs.setdefault(sp,[]).append(parsearch(arch))

    def add_event(self,edge,event):
        self.events.setdefault(edge,[]).append(parseevent(event))

    def add_cost(self,edge,cost):
        try:
            float(cost)
        except ValueError:
            raise ValueError('Value not a valid number: ' + cost)
        if edge in self.costs:
            raise Exception('Key already found in costs: ' + edge)
        self.costs[edge] = cost


    def read_data(self,filename,convert_names=True):
        state = self.EMPTY
        for line in util.open_stream(filename,'r'):
            line = line.rstrip()

            # determine current state
            oldstate = state
            changedState = False
            for s in self.__statestr.iterkeys():
                if line == self.__startstr(s):
                    state = s
                    changedState = True
                    break
            if not changedState:
                for s in self.__statestr.iterkeys():
                    if line == self.__endstr(s):
                        if s == self.ARCHTREE:
                            return   # end of arch tree
                        state = self.ARCHTREE
                        changedState = True
                        break
                                        
            # perform action
            if changedState:
                # check state transition
                assert(self.__statetrans(oldstate,state)),(oldstate,state)
                continue
            elif state == self.EMPTY:
                continue
            elif state == self.ARCHTREE:
                continue
            elif state == self.INPUTS:
                tokens = line.split('=')
                if ':' in line: # deprecated
                    tokens = line.split(':')
                self.add_input(tokens[0],float(tokens[1]))
            elif state == self.TIED:
                tokens = line.split('=')
                if tokens[1] == 'T':
                    self.add_tied(tokens[0],True)
                elif tokens[1] == 'F':
                    self.add_tied(tokens[0],False)
            elif state == self.ARCHS:
                # gene and architecture by species
                tokens = line.split('\t')
                sp,gnarchs = tokens[0],tokens[1:]
                if convert_names and sp.isdigit():
                    sp = int(sp)
                if sp in self.archs:
                    raise Exception("Species already in list: " + sp)
                self.archs[sp] = []
                for a in gnarchs:
                    self.add_arch(sp,a)
            elif state == self.EVENTS:
                # events by edge
                tokens = line.split('\t')
                edge = parseedge(tokens[0])
                if convert_names:
                    if edge[0].isdigit():
                        parent = int(edge[0])
                    else:
                        parent = edge[0]
                    if edge[1].isdigit():
                        child = int(edge[1])
                    else:
                        child = edge[1]
                    edge = (parent,child)
                if edge in self.events:
                    raise Exception("Edge already in events list: " + line)
                self.events[edge] = []
                for e in tokens[1:]:
                    self.add_event(edge,e)
            elif state == self.COSTS:
                # costs by edge
                tokens = line.split('\t')
                if tokens[0] == 'TOTAL':
                    edge = 'TOTAL'
                else:
                    edge = parseedge(tokens[0])
                self.add_cost(edge,float(tokens[1]))
                
        # no tree found
        if state == self.EMPTY:
            raise Exception("No arch tree found")

        
    def write_data(self,filename='-'):
        opened = False if isinstance(filename,file) else True
        out = util.open_stream(filename,'w')

        print >>out,"# ARCHTREE\n"
        if len(self.inputs) > 0:
            print >>out,self.__startstr(self.INPUTS)
            for key,val in self.inputs.iteritems():
                print >>out,'%s=%s' % (key,val)
            print >>out,self.__endstr(self.INPUTS)+'\n'
        if len(self.tied) > 0:
            print >>out,self.__startstr(self.TIED)
            for key,val in self.tied.iteritems():
                print >>out,'%s=%s' % (key,'T' if val else 'F')
            print >>out,self.__endstr(self.TIED)+'\n'
        if len(self.archs) > 0:
            print >>out,self.__startstr(self.ARCHS)
            for sp,gnarchs in self.archs.iteritems():
                out.write(str(sp))
                for a in gnarchs:
                    out.write('\t%s' % stringarch(a))
                out.write('\n')
            print >>out,self.__endstr(self.ARCHS)+'\n'
        if len(self.events) > 0:
            print >>out,self.__startstr(self.EVENTS)
            for edge,events in self.events.iteritems():
                out.write('(%s,%s)' % (str(edge[0]),str(edge[1])))
                for e in events:
                    out.write('\t%s' % stringevent(e))
                out.write('\n')
            print >>out,self.__endstr(self.EVENTS)+'\n'
        if len(self.costs) > 0:
            print >>out,self.__startstr(self.COSTS)
            for edge,cost in self.costs.iteritems():
                print >>out,'%s\t%s' % (edge if edge=='TOTAL' \
                                        else '(%s,%s)' % (edge[0],edge[1]),
                                        cost)
            print >>out,self.__endstr(self.COSTS)+'\n'
        print >>out, "# ENDARCHTREE\n\n"
            
        if opened:
            out.close()
    writeData = write_data

    def get_regions(self):
        regs = []
        for sp,gnarchs in self.archs.iteritems():
            for a in gnarchs:
                regs.extend(a[1].split('-'))
        return regs

#============================================================================
# I/O utilities
def read_archfam(filename,convert_names=True):
    archfam = ArchFam()
    archfam.read_data(filename,convert_names)
    return archfam

def iter_archfams(filename,convert_names=True):
    # iterate through multiple archfams
    infile = open(filename)
    try:
        while True:
            yield read_archfam(infile,convert_names)
    except:
        pass
    finally:
        infile.close()

def read_archfams(filename,convert_names=True):
    return list(iter_archfams)

def iter_archfamfiles(args,treeext='.mp'):
    # iterate through archfamfiles
    for archfamfile in args:
        archfamfile = archfamfile.rstrip("\n")
        famid = int(util.replace_ext(os.path.basename(archfamfile),treeext,''))
       
        if not os.path.exists(archfamfile):
            print >>sys.stderr,'famid %s does not have a %s file' % (famid,treeext)
            continue
        yield famid,list(iter_archfams(archfamfile))

def read_archfamfiles(args,treeext='.mp'):
    return list(iter_archfamfiles)

def iter_archfam_boot(args,treeext='.mp',Boot=1,boot=1):
    # iterate through bootstraps
    cache1 = {}
    cache2 = {}
    numargs = len(args)
    rndvec = np.asarray(np.floor(np.random.rand(numargs,Boot)*boot),dtype=int)
    numtrees = boot

    # get data
    for filendx,(famid,archfams) in enumerate(iter_archfamfiles(args,treeext)):
        assert(boot==len(archfams)),(famid,boot,len(archfams))
        cache1[filendx] = famid
        for myboot,archfam in enumerate(archfams):
            if myboot in rndvec[filendx,:]:
                cache2[filendx,myboot] = archfam
    
    # generator
    for myBoot in xrange(Boot):
        for filendx in xrange(numargs):
            # famid, Boot, boot, archfam
            yield cache1[filendx],myBoot,rndvec[filendx,myBoot],cache2[filendx,rndvec[filendx,myBoot]]

def read_archfam_boot(args,treeext='.mp',Boot=1,boot=1):
    return list(iter_archfam_boot)

#============================================================================
# counting
#
def init_event_tree(stree):
    # initialize counts to zero
    def walk(node):
        node.data['genes'] = 0
        node.data['regions'] = 0
        node.data['appear'] = 0
        node.data['dup'] = 0
        node.data['loss'] = 0
        node.data['merge'] = 0
        node.data['split'] = 0
        node.recurse(walk)
    walk(stree.root)
    
def count_all_tree(stree,archfam):
    stree = count_genes_tree(stree,archfam)
    stree = count_regions_tree(stree,archfam)
    stree = count_events_tree(stree,archfam)
    return stree

def count_genes_tree(stree,archfam):
    for sp,archs in archfam.archs.iteritems():
        stree.nodes[sp].data['genes'] += len(archs)
    return stree

def count_regions_tree(stree,archfam):
    for sp,archs in archfam.archs.iteritems():
        for gn,arch in archs:
            stree.nodes[sp].data['regions'] += len(arch.split('-'))        
    return stree

def count_events_tree(stree,archfam):
    for (parent,child),lst in archfam.events.iteritems():
        for etype,estart,eend in lst:
            if etype == 'G':
                stree.nodes[child].data['appear'] += 1
            elif etype == 'D':
                stree.nodes[child].data['dup'] += 1
            elif etype == 'L':
                stree.nodes[child].data['loss'] += 1
            elif etype == 'M':
                stree.nodes[child].data['merge'] += 1
            elif etype == 'S':
                stree.nodes[child].data['split'] += 1
            else:
                raise Exception("Invalid event type: %s" % etype)
    return stree
