import sys,os
import collections

from rasmus import svg
from rasmus import util
from rasmus import treelib
from rasmus.vis import treesvg

def draw_events_tree(tree, events,
                     mergeColor=(0,0,1), splitColor=(1,0,0), size=4,
                     labels={}, xscale=100, yscale=20, canvas=None,
                     leafPadding=10,
                     labelOffset=None, fontSize=10, labelSize=None,
                     minlen=1, maxlen=util.INF, filename=sys.stdout,
                     rmargin=150, lmargin=10, tmargin=0, bmargin=None,
                     colormap=None,
                     legendScale=False, autoclose=None):
    # set defaults (copied from treesvg)
    if labelOffset == None:
        labelOffset = -1

    if bmargin == None:
        bmargin = yscale
    
    if sum(x.dist for x in tree.nodes.values()) == 0:
        legendScale = False
        minlen = xscale

    # layout tree (copied from treesvg)
    coords = treelib.layout_tree(tree, xscale, yscale, minlen, maxlen)

    xcoords, ycoords = zip(* coords.values())
    maxwidth = max(xcoords)
    maxheight = max(ycoords) + labelOffset

    # initialize canvas (copied from treesvg)
    if canvas == None:
        canvas = svg.Svg(util.open_stream(filename, "w"))
        width = int(rmargin + maxwidth + lmargin )
        height = int(tmargin + maxheight + bmargin)
        
        canvas.beginSvg(width, height)
        
        if autoclose == None:
            autoclose = True
    else:
        if autoclose == None:
            autoclose = False
            
    # draw tree with layout (delegate to treesvg)
    treesvg.draw_tree(tree, labels=labels, xscale=xscale, yscale=yscale, canvas=canvas,
              leafPadding=leafPadding,
              labelOffset=labelOffset, fontSize=fontSize, labelSize=labelSize,
              minlen=minlen, maxlen=maxlen, filename=filename,
              colormap=colormap,
              layout=coords,
              legendScale=legendScale, autoclose=False)

    canvas.beginTransform(("translate", lmargin, tmargin))
    
    # draw histogram of events
    draw_events_hist(events,canvas,tree,coords,
                     mergeColor,splitColor,size)

    canvas.endTransform()

    # close canvas (copied from treesvg)
    if autoclose:
        canvas.endSvg()

    return canvas
drawEventsTree = draw_events_tree


def draw_events_hist(events,canvas,tree,coords,
                     mergeColor=(0,0,1),
                     splitColor=(1,0,0),
                     size=4):
    maxct = 0
    for d in events.itervalues():
        maxct = max(maxct,max(d.values()))
    maxct = float(maxct)

    # get scale
    k = util.INF
    for node in tree:
        if node.parent:
            if node.name in events:
                branchlen = coords[node][0]-coords[node.parent][0]
                ct = max(events[node.name].values())
                k = min(k,branchlen)
    k = float(k)
                
    # draw
    def walk(node):
        x,y = coords[node]
        if node.parent:
            parentx = coords[node.parent][0]
            branchlen = x - parentx

            if node.name in events:
                # draw merges
                ct = events[node.name]["merge"]
                canvas.rect(parentx,y-size,
                            k*ct/maxct,size,fillColor=mergeColor,strokeColor=(0,0,0,0))
                canvas.text(str(ct),parentx+.1*branchlen,y-1.2*size,size,fillColor=mergeColor)
                
                # draw splits
                ct = events[node.name]["split"]
                canvas.rect(parentx,y,
                            k*ct/maxct,size,fillColor=splitColor,strokeColor=(0,0,0,0))
                canvas.text(str(ct),parentx+.1*branchlen,y+2.2*size,size,fillColor=splitColor)

        if not node.is_leaf():
            for child in node.children:
                walk(child)
    walk(tree.root)        
drawEventsHist = draw_events_hist
    

def draw_arch_tree(tree, regs, gnarchs, events, boxSize=4, eventSize=4,
                   regs_colormap=None,
                   labels={}, xscale=100, yscale=20, canvas=None,
                   leafPadding=10,
                   labelOffset=None, fontSize=10, labelSize=None,
                   minlen=1, maxlen=util.INF, filename=sys.stdout,
                   rmargin=150, lmargin=10, tmargin=0, bmargin=None,
                   colormap=None,
                   legendScale=False, autoclose=None,
                   legendWidth=80, legendSize=10, annotateregs=True):

    # make sure regs are unique
    regs = sorted(set(regs))

    # set defaults (copied from treesvg)
    if labelOffset == None:
        labelOffset = -1

    if bmargin == None:
        bmargin = yscale
    
    if sum(x.dist for x in tree.nodes.values()) == 0:
        legendScale = False
        minlen = xscale

    # layout tree (copied from treesvg)
    coords = treelib.layout_tree(tree, xscale, yscale, minlen, maxlen)

    xcoords, ycoords = zip(* coords.values())
    maxwidth = max(xcoords)
    maxheight = max(ycoords) + labelOffset

    # initialize canvas (copied from treesvg)
    if canvas == None:
        canvas = svg.Svg(util.open_stream(filename, "w"))
        width = int(rmargin + maxwidth + lmargin + legendWidth)
        height = int(tmargin + maxheight + bmargin)
        
        canvas.beginSvg(width, height)
        
        if autoclose == None:
            autoclose = True
    else:
        if autoclose == None:
            autoclose = False
    
    # draw tree with layout (delegate to treesvg)
    treesvg.draw_tree(tree, labels=labels, xscale=xscale, yscale=yscale, canvas=canvas,
              leafPadding=leafPadding,
              labelOffset=labelOffset, fontSize=fontSize, labelSize=labelSize,
              minlen=minlen, maxlen=maxlen, filename=filename,
              colormap=colormap,
              layout=coords,
              legendScale=legendScale, autoclose=False)

    canvas.beginTransform(("translate", lmargin, tmargin))
    
    # draw archs and events
    if regs_colormap is None:
        regs_colormap = __get_reg_cmap(regs)
    draw_archs(gnarchs,canvas,tree,coords,
               cmap=regs_colormap,boxSize=boxSize)
    if events is not None:
        draw_events(events,canvas,tree,coords,
                    cmap=regs_colormap,eventSize=eventSize)

    # annotate
    if annotateregs:
        boxSize = legendSize
        fontSize = legendSize
        offset = maxheight/2.-(len(regs)*1.5*boxSize)/2.
        for r in regs:
            x = maxwidth+rmargin
            canvas.rect(x,offset,
                        boxSize,boxSize,fillColor=regs_colormap[r])
            canvas.text(str(r),x+1.5*boxSize,
                        offset+fontSize,fontSize)
            offset += 1.5*boxSize

    canvas.endTransform()

    # close canvas (copied from treesvg)
    if autoclose:
        canvas.endSvg()

    return canvas
drawArchTree = draw_arch_tree


def __get_reg_cmap(regs):
    # get colors for each region
    rcmap = util.rainbowColorMap(low=0,high=len(regs)-1)
    cmap = dict([(r,rcmap.get(ndx)) for ndx,r in enumerate(regs)])
    return cmap


# helper function to draw an architecture
def __draw_arch(arch, canvas, x, y, cmap, boxSize):
    for j,r in enumerate(arch.split('-')):
        canvas.rect(x+j*boxSize,y,
                    boxSize,boxSize,fillColor=cmap[r])


def draw_archs(gnarchs, canvas, tree, coords, cmap, boxSize=4):
    # get maximum number of regions in an architecture
    # and get counts for each architecture
    maxlen = 0
    archs = {}
    for name,node in tree.nodes.iteritems():
        archs[node] = collections.defaultdict(list)
        for gene,arch in gnarchs.get(name,[]):
            if len(arch) == 0:  # do not draw null architectures
                continue
            maxlen = max(maxlen,len(arch.split('-')))
            archs[node][arch].append(gene)
    
    # draw archs
    J = maxlen+2  # xoffset
    def walk(node):
        x,y = coords[node]
        I = len(archs[node])-1  # yoffset
        for i,(arch,genes) in enumerate(sorted(archs[node].items(),key=lambda (arch,genes):arch)):
            ydraw = y+(i-I)*1.5*boxSize-0.5*boxSize
            __draw_arch(arch,canvas,x-J*boxSize,ydraw-boxSize,cmap,boxSize)

            # write count if > 1
            ct = len(genes)
            if ct > 1:
                canvas.text(str(ct),x-boxSize,ydraw,boxSize)

            # write gene name if specified
            line = ','.join([gene for gene in genes if gene is not None])
            if line != '':
                canvas.text(line,x,ydraw,boxSize)

        if not node.is_leaf():
            for child in node.children:
                walk(child)
    walk(tree.root)


def draw_events(events, canvas, tree, coords, cmap, eventSize=4):
    # get counts for each event
    eventcts = {}
    for name,node in tree.nodes.iteritems():
        if cmp(node,tree.root) == 0:
            continue
        ct = {}
        for t,s,e in events.get((node.parent.name,name),[]):
            ct[t,s,e] = ct.get((t,s,e),0) + 1
        eventcts[node.parent,node] = sorted(ct.items())

    # draw events
    numregs = len(cmap)
    def walk(node):
        x,y = coords[node]
        if node.parent:
            parentx = coords[node.parent][0]
            branchlen = x - parentx
            I = len(eventcts[node.parent,node])-1  # yoffset
            for i,((t,s,e),ct) in enumerate(eventcts[node.parent,node]):
                ydraw = y+(i-I)*1.5*eventSize-0.5*eventSize
                xoffset = parentx+0.05*branchlen

                canvas.text(t,xoffset,ydraw,eventSize)
                xoffset += 1.5*eventSize

                # draw starting architectures
                for arch in s.split(','):
                    if len(arch) == 0:  # do not draw null architectures
                        continue
                    __draw_arch(arch,canvas,
                                xoffset,ydraw-eventSize,cmap,eventSize)
                    xoffset += (len(arch.split('-'))+1)*eventSize
                    
                # TODO: convert to arrow
                canvas.text('>',xoffset,ydraw,eventSize)
                xoffset += 1.2*eventSize

                # draw ending architectures
                for arch in e.split(','):
                    if len(arch) == 0:
                        continue
                    __draw_arch(arch,canvas,
                                xoffset,ydraw-eventSize,cmap,eventSize)
                    xoffset += (len(arch.split('-'))+1)*eventSize

                # write count if > 1
                if ct > 1:
                    canvas.text(str(ct),xoffset,ydraw,eventSize)

        if not node.is_leaf():
            for child in node.children:
                walk(child)
    walk(tree.root)      


# copied from treesvg
def show_events_tree(tree, events, *args, **kargs):
    if not os.fork():
        kargs.setdefault('filename', os.popen("display", "w"))
        kargs.setdefault('legendScale', True)
        draw_events_tree(tree, events, *args, **kargs)
        sys.exit(0)
showEventsTree = show_events_tree

def show_arch_tree(tree, regs, gnarchs, events, *args, **kargs):
    if not os.fork():
        kargs.setdefault('filename', os.popen("display", "w"))
        kargs.setdefault('legendScale', True)
        draw_arch_tree(tree, regs, gnarchs, events, *args, **kargs)
        sys.exit(0)
showArchTree = show_arch_tree
